@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 # __init__.py
 
-# Copyright (c) 2016-2019, Richard Gerum
+# Copyright (c) 2016-2020, Richard Gerum
 #
 # This file is part of Pylustrator.
 #
@@ -19,12 +19,10 @@
 # You should have received a copy of the GNU General Public License
 # along with Pylustrator. If not, see <http://www.gnu.org/licenses/>
 
-from .drag_bib import StartPylustrator, moveArtist
 from .QtGuiDrag import initialize as start
 from .helper_functions import fig_text, add_axes, add_image, despine, changeFigureSize, mark_inset, VoronoiPlot, selectRectangle, mark_inset_pos, draw_from_point_to_bbox, draw_from_point_to_point, loadFigureFromFile, add_letter, add_letters
 from .QtGui import initialize as StartColorChooser
 from .lab_colormap import LabColormap
-from .jupyter_cells import setCellInput
 from .helper_functions import loadFigureFromFile as load
 
-__version__ = '0.11.0'
+__version__ = '1.0.0'
