@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 # change_tracker.py
 
-# Copyright (c) 2016-2019, Richard Gerum
+# Copyright (c) 2016-2020, Richard Gerum
 #
 # This file is part of Pylustrator.
 #
@@ -19,17 +19,18 @@
 # You should have received a copy of the GNU General Public License
 # along with Pylustrator. If not, see <http://www.gnu.org/licenses/>
 
-from __future__ import division, print_function
-
 import re
 import sys
 import traceback
+from typing import IO
 
 import matplotlib
 import matplotlib as mpl
 import matplotlib.pyplot as plt
 from matplotlib import _pylab_helpers
+from matplotlib.artist import Artist
 from matplotlib.axes._subplots import Axes
+from matplotlib.collections import Collection
 from matplotlib.figure import Figure
 from matplotlib.lines import Line2D
 from matplotlib.patches import Rectangle
@@ -40,7 +41,8 @@ from .exception_swallower import Dummy
 from .jupyter_cells import open
 
 
-def getReference(element, allow_using_variable_names=True):
+def getReference(element: Artist, allow_using_variable_names=True):
+    """ get the code string that represents the given Artist. """
     if element is None:
         return ""
     if isinstance(element, Figure):
@@ -55,6 +57,9 @@ def getReference(element, allow_using_variable_names=True):
     if isinstance(element, matplotlib.lines.Line2D):
         index = element.axes.lines.index(element)
         return getReference(element.axes) + ".lines[%d]" % index
+    if isinstance(element, matplotlib.collections.Collection):
+        index = element.axes.collections.index(element)
+        return getReference(element.axes) + ".collections[%d]" % index
     if isinstance(element, matplotlib.patches.Patch):
         if element.axes:
             index = element.axes.patches.index(element)
@@ -117,7 +122,8 @@ def getReference(element, allow_using_variable_names=True):
     raise TypeError(str(type(element)) + " not found")
 
 
-def setFigureVariableNames(figure):
+def setFigureVariableNames(figure: Figure):
+    """ get the global variable names that refer to the given figure """
     import inspect
     mpl_figure = _pylab_helpers.Gcf.figs[figure].canvas.figure
     calling_globals = inspect.stack()[2][0].f_globals
@@ -133,10 +139,11 @@ def setFigureVariableNames(figure):
 
 
 class ChangeTracker:
+    """ a class that records a list of the change to the figure """
     changes = None
     saved = True
 
-    def __init__(self, figure):
+    def __init__(self, figure: Figure):
         global stack_position
         self.figure = figure
         self.edits = []
@@ -155,7 +162,8 @@ class ChangeTracker:
 
         self.load()
 
-    def addChange(self, command_obj, command, reference_obj=None, reference_command=None):
+    def addChange(self, command_obj: Artist, command: str, reference_obj: Artist = None, reference_command: str = None):
+        """ add a change """
         command = command.replace("\n", "\\n")
         if reference_obj is None:
             reference_obj = command_obj
@@ -164,7 +172,8 @@ class ChangeTracker:
         self.changes[reference_obj, reference_command] = (command_obj, command)
         self.saved = False
 
-    def removeElement(self, element):
+    def removeElement(self, element: Artist):
+        """ remove an Artis from the figure """
         # create_key = key+".new"
         created_by_pylustrator = (element, ".new") in self.changes
         # delete changes related to this element
@@ -179,13 +188,15 @@ class ChangeTracker:
             element.remove()
         self.figure.selection.remove_target(element)
 
-    def addEdit(self, edit):
+    def addEdit(self, edit: list):
+        """ add an edit to the stored list of edits """
         if self.last_edit < len(self.edits) - 1:
             self.edits = self.edits[:self.last_edit + 1]
         self.edits.append(edit)
         self.last_edit = len(self.edits) - 1
 
     def backEdit(self):
+        """ undo an edit in the list """
         if self.last_edit < 0:
             return
         edit = self.edits[self.last_edit]
@@ -194,6 +205,7 @@ class ChangeTracker:
         self.figure.canvas.draw()
 
     def forwardEdit(self):
+        """ redo an edit """
         if self.last_edit >= len(self.edits) - 1:
             return
         edit = self.edits[self.last_edit + 1]
@@ -202,6 +214,7 @@ class ChangeTracker:
         self.figure.canvas.draw()
 
     def load(self):
+        """ load a set of changes from a script file. The changes are the code that pylustrator generated """
         regex = re.compile(r"(\.[^\(= ]*)(.*)")
         command_obj_regexes = [getReference(self.figure),
                                r"plt\.figure\([^)]*\)",
@@ -211,8 +224,10 @@ class ChangeTracker:
                                r"\.texts\[\d*\]",
                                r"\.{title|_left_title|_right_title}",
                                r"\.lines\[\d*\]",
+                               r"\.collections\[\d*\]",
                                r"\.patches\[\d*\]",
                                r"\.get_[xy]axis\(\)\.get_(major|minor)_ticks\(\)\[\d*\]",
+                               r"\.get_[xy]axis\(\)\.get_label\(\)",
                                r"\.get_legend\(\)",
                                ]
         command_obj_regexes = [re.compile(r) for r in command_obj_regexes]
@@ -262,7 +277,7 @@ class ChangeTracker:
             reference_command = command
 
             if command == ".set_xticks" or command == ".set_yticks" or command == ".set_xlabels" or command == ".set_ylabels":
-                if line.find("minor=True"):
+                if line.find("minor=True") != -1:
                     reference_command = command + "_minor"
 
             # for new created texts, the reference object is the text and not the axis/figure
@@ -281,38 +296,42 @@ class ChangeTracker:
 
             self.get_reference_cached[reference_obj] = reference_obj_str
 
+            #print("---", [reference_obj, reference_command], (command_obj, command + parameter))
             self.changes[reference_obj, reference_command] = (command_obj, command + parameter)
         self.sorted_changes()
 
     def sorted_changes(self):
+        """ sort the changes by their priority. For example setting to logscale needs to be executed before xlim. """
         def getRef(obj):
             try:
                 return getReference(obj)
             except (ValueError, TypeError):
-                return ""
-            if obj in self.get_reference_cached:
-                return self.get_reference_cached[obj]
-            else:
-                try:
-                    return getReference(obj)
-                except (ValueError, TypeError):
-                    return ""
+                # the ticks objects can for some reason not be referenced properly in the next code run
+                # they are somehow XTicks objects when loading but when saving they are Text objects
+                if obj in self.get_reference_cached:
+                    return self.get_reference_cached[obj]
+                raise
 
         indices = []
         for reference_obj, reference_command in self.changes:
-            if isinstance(reference_obj, Figure):
-                obj_indices = ("", "", "", "")
-            else:
-                if getattr(reference_obj, "axes", None) is not None:
-                    if reference_command == ".new":
-                        index = "0"
-                    else:
-                        index = "1"
-                    obj_indices = (getRef(reference_obj.axes), getRef(reference_obj), index, reference_command)
+            try:
+                if isinstance(reference_obj, Figure):
+                    obj_indices = ("", "", "", "")
                 else:
-                    obj_indices = (getRef(reference_obj), "", "", reference_command)
-            indices.append(
-                [(reference_obj, reference_command), self.changes[reference_obj, reference_command], obj_indices])
+                    if getattr(reference_obj, "axes", None) is not None:
+                        if reference_command == ".new":
+                            index = "0"
+                        elif reference_command == ".set_xscale" or reference_command == ".set_yscale":
+                            index = "1"
+                        else:
+                            index = "2"
+                        obj_indices = (getRef(reference_obj.axes), getRef(reference_obj), index, reference_command)
+                    else:
+                        obj_indices = (getRef(reference_obj), "", "", reference_command)
+                indices.append(
+                    [(reference_obj, reference_command), self.changes[reference_obj, reference_command], obj_indices])
+            except (ValueError, TypeError) as err:
+                print(err, file=sys.stderr)
 
         srt = natsorted(indices, key=lambda a: a[2])
         output = []
@@ -321,11 +340,12 @@ class ChangeTracker:
             try:
                 output.append(getRef(command_obj) + command)
             except TypeError as err:
-                print(err)
+                print(err, file=sys.stderr)
 
         return output
 
     def save(self):
+        """ save the changes to the .py file """
         header = [getReference(self.figure) + ".ax_dict = {ax.get_label(): ax for ax in " + getReference(
             self.figure) + ".axes}", "import matplotlib as mpl"]
 
@@ -351,7 +371,8 @@ class ChangeTracker:
         self.saved = True
 
 
-def getTextFromFile(block_id, stack_pos):
+def getTextFromFile(block_id: str, stack_pos: traceback.FrameSummary):
+    """ get the text which corresponds to the block_id (e.g. which figure) at the given position sepcified by stack_pos. """
     block_id = lineToId(block_id)
     block = None
 
@@ -383,28 +404,34 @@ def getTextFromFile(block_id, stack_pos):
 
 
 class Block:
+    """ an object to represent the code block generated by a pylustrator save """
     id = None
     finished = False
 
-    def __init__(self, line):
+    def __init__(self, line: str):
+        """ initialize the block with its first line """
         self.text = line
         self.size = 1
         self.indent = getIndent(line)
 
-    def add(self, line):
+    def add(self, line: str):
+        """ add a line to the block """
         if self.id is None:
             self.id = lineToId(line)
         self.text += line
         self.size += 1
 
     def end(self):
+        """ end the block """
         self.finished = True
 
     def __iter__(self):
+        """ iterate over all the lines of the block """
         return iter(self.text.split("\n"))
 
 
-def getIndent(line):
+def getIndent(line: str):
+    """ get the indent part of a line of code """
     i = 0
     for i in range(len(line)):
         if line[i] != " " and line[i] != "\t":
@@ -413,18 +440,20 @@ def getIndent(line):
     return indent
 
 
-def addLineCounter(fp):
+def addLineCounter(fp: IO):
+    """ wrap a file pointer to store th line numbers """
     fp.lineno = 0
     write = fp.write
 
-    def write_with_linenumbers(line):
+    def write_with_linenumbers(line: str):
         write(line)
         fp.lineno += line.count("\n")
 
     fp.write = write_with_linenumbers
 
 
-def lineToId(line):
+def lineToId(line: str):
+    """ get the id of a line, e.g. part which specifies which figure it refers to """
     line = line.strip()
     line = line.split(".ax_dict")[0]
     if line.startswith("fig = "):
@@ -432,7 +461,8 @@ def lineToId(line):
     return line
 
 
-def insertTextToFile(new_block, stack_pos, figure_id_line):
+def insertTextToFile(new_block: str, stack_pos: traceback.FrameSummary, figure_id_line: str):
+    """ insert a text block into a file """
     figure_id_line = lineToId(figure_id_line)
     block = None
     written = False
